import {
  HASHTAG_FILTER,
  MASTODON_INSTANCE,
  ODESLI_API_KEY,
  URL_FILTER,
  YOUTUBE_API_KEY,
  YOUTUBE_DISABLE
} from '$env/static/private';
import { log } from '$lib/log';
import type { Post, Tag, TimelineEvent } from '$lib/mastodon/response';
import type { OdesliResponse, Platform, SongInfo } from '$lib/odesliResponse';
import { getPosts, savePost } from '$lib/server/db';
import { createFeed, saveAtomFeed } from '$lib/server/rss';
import { sleep } from '$lib/sleep';
import { isTruthy } from '$lib/truthyString';
import { WebSocket } from 'ws';

const YOUTUBE_REGEX = new RegExp(
  /https?:\/\/(www\.)?youtu((be.com\/.*?v=)|(\.be\/))(?<videoId>[a-zA-Z_0-9-]+)/gm
);

const URL_REGEX = new RegExp(/href="(?<postUrl>[^>]+?)" target="_blank"/gm);

export class TimelineReader {
  private static _instance: TimelineReader;

  private static async isMusicVideo(videoId: string) {
    if (!YOUTUBE_API_KEY || YOUTUBE_API_KEY === 'CHANGE_ME') {
      // Assume that it *is* a music link when no YT API key is provided
      // If it should assumed to not be YOUTUBE_DISABLE needs to be set to something truthy
      return true;
    }
    const searchParams = new URLSearchParams([
      ['part', 'snippet'],
      ['id', videoId],
      ['key', YOUTUBE_API_KEY]
    ]);
    const youtubeVideoUrl = new URL(`https://www.googleapis.com/youtube/v3/videos?${searchParams}`);
    const resp = await fetch(youtubeVideoUrl);
    const respObj = await resp.json();
    if (!respObj.items.length) {
      log.warn('Could not find video with id', videoId);
      return false;
    }

    const item = respObj.items[0];
    if (item.tags?.includes('music')) {
      return true;
    }

    const categorySearchParams = new URLSearchParams([
      ['part', 'snippet'],
      ['id', item.categoryId],
      ['key', YOUTUBE_API_KEY]
    ]);
    const youtubeCategoryUrl = new URL(
      `https://www.googleapis.com/youtube/v3/videoCategories?${categorySearchParams}`
    );
    const categoryTitle: string = await fetch(youtubeCategoryUrl)
      .then((r) => r.json())
      .then((r) => r.items[0]?.title);
    return categoryTitle === 'Music';
  }

  private static async checkYoutubeMatches(postContent: string): Promise<string | null> {
    if (isTruthy(YOUTUBE_DISABLE)) {
      return null;
    }
    const matches = postContent.matchAll(YOUTUBE_REGEX);
    for (const match of matches) {
      if (match === undefined || match.groups === undefined) {
        continue;
      }
      const videoId = match.groups.videoId.toString();
      try {
        const isMusic = await TimelineReader.isMusicVideo(videoId);
        if (isMusic) {
          return match[0];
        }
      } catch (e) {
        log.error('Could not check if', videoId, 'is a music video', e);
      }
    }
    return null;
  }

  private static async getSongInfo(url: URL, remainingTries = 6): Promise<SongInfo | null> {
    if (remainingTries === 0) {
      log.error('No tries remaining. Lookup failed!');
      return null;
    }
    if (url.hostname === 'songwhip.com') {
      // song.link doesn't support songwhip links and songwhip themselves will provide metadata if you pass in a
      // Apple Music/Spotify/etc link, but won't when provided with their own link, so no way to extract song info
      // except maybe scraping their HTML
      return null;
    }

    const odesliParams = new URLSearchParams();
    odesliParams.append('url', url.toString());
    odesliParams.append('userCountry', 'DE');
    odesliParams.append('songIfSingle', 'true');
    if (ODESLI_API_KEY && ODESLI_API_KEY !== 'CHANGE_ME') {
      odesliParams.append('key', ODESLI_API_KEY);
    }
    const odesliApiUrl = `https://api.song.link/v1-alpha.1/links?${odesliParams}`;
    try {
      return fetch(odesliApiUrl).then(async (response) => {
        if (response.status === 429) {
          throw new Error('Rate limit reached', { cause: 429 });
        }
        return response.json().then((odesliInfo: OdesliResponse) => {
          const info = odesliInfo.entitiesByUniqueId[odesliInfo.entityUniqueId];
          const platform: Platform = 'youtube';
          return {
            ...info,
            pageUrl: odesliInfo.pageUrl,
            youtubeUrl: odesliInfo.linksByPlatform[platform]?.url,
            postedUrl: url.toString()
          } as SongInfo;
        });
      });
    } catch (e) {
      if (e instanceof Error && e.cause === 429) {
        log.warn('song.link rate limit reached. Trying again in 10 seconds');
        await sleep(10_000);
        return await this.getSongInfo(url, remainingTries - 1);
      }
      log.error(`Failed to load ${url} info from song.link`, e);
      return null;
    }
  }

  private static async getUrlFromPreviewCard(post: Post): Promise<string | undefined> {
    return undefined;
    // Currently disabled, because it seems to always be null, even after re-fetching the post from Mastodon
    /*
    if (post.card) {
      return post.card?.url;
    }
    try {
      const status: Post = await (
        await fetch(`https://${MASTODON_INSTANCE}/api/v1/statuses/${post.id}`)
      ).json();
      return status.card?.url;
    } catch (e) {
      log.error(`Could not fetch status ${post.url}`, e);
    }
    */
  }

  private startWebsocket() {
    const socket = new WebSocket(`wss://${MASTODON_INSTANCE}/api/v1/streaming`);
    socket.onopen = () => {
      log.log('Connected to WS');
      socket.send('{ "type": "subscribe", "stream": "public:local"}');
    };
    socket.onmessage = async (event) => {
      try {
        const data: TimelineEvent = JSON.parse(event.data.toString());
        if (data.event !== 'update') {
          return;
        }
        const post: Post = JSON.parse(data.payload);
        const hashttags: string[] = HASHTAG_FILTER.split(',');
        const found_tags: Tag[] = post.tags.filter((t: Tag) => hashttags.includes(t.name));

        const urls: string[] = URL_FILTER.split(',');
        const found_urls = urls.filter((t) => post.content.includes(t));
        // If we don't have any tags or non-youtube urls, check youtube
        // YT is handled separately, because it requires an API call and therefore is slower
        if (found_urls.length === 0 && found_tags.length === 0) {
          const youtubeUrl = await TimelineReader.checkYoutubeMatches(post.content);
          if (youtubeUrl === null) {
            log.log('Ignoring post', post.url);
            return;
          }
          log.debug('Found YT URL', youtubeUrl, found_urls, found_urls.length);
        }

        // TODO: Change URL detection above to use this regex.
        // Looks like we're stuck with regex for now instead of using preview cards.
        // Might as well use it to find URLs. Could also use this for YouTube: If Odesli finds something, it's a song,
        // if not, ignore it. No need to consult the YT API and give those links a special handling
        const musicUrls: URL[] = [];
        const musicUrl = await TimelineReader.getUrlFromPreviewCard(post);
        if (musicUrl) {
          try {
            musicUrls.push(new URL(musicUrl));
          } catch (e) {
            log.error(
              'URL received from preview card does not seem to be a valid URL',
              musicUrl,
              e
            );
          }
        } else {
          const urlMatches = post.content.matchAll(URL_REGEX);
          for (const match of urlMatches) {
            if (match === undefined || match.groups === undefined) {
              console.warn(
                'Match listed in allMatches, but either it or its groups are undefined',
                match
              );
              continue;
            }
            const urlMatch = match.groups.postUrl.toString();
            let url: URL;
            try {
              url = new URL(urlMatch);
            } catch (e) {
              log.error('URL found via Regex does not seem to be a valud url', urlMatch, e);
              continue;
            }

            // Check *all* found url and let odesli determine if it is music or not
            musicUrls.push(url);
          }
        }

        const songs: SongInfo[] = [];
        log.debug(`Checking ${musicUrls.length} URLs if they contain song data`);
        for (const url of musicUrls) {
          let hostname: string | null = null;
          try {
            hostname = new URL(url).hostname;
          } catch (e) {
            log.error(`Could not check hostname for URL ${url}`, e);
          }
          if (hostname === 'songwhip.com') {
            // TODO: Implement checking the songwhip API
            continue;
          }
          const info = await TimelineReader.getSongInfo(url);
          log.debug(`Found song info for ${url}?`, info);
          if (info) {
            songs.push(info);
          }
        }

        await savePost(post, songs);
        log.debug('Saved post', post.url);

        const posts = await getPosts(null, null, 100);
        await saveAtomFeed(createFeed(posts));
      } catch (e) {
        log.error('error message', event, event.data, e);
      }
    };
    socket.onclose = (event) => {
      log.warn(
        `Websocket connection to ${MASTODON_INSTANCE} closed. Code: ${event.code}, reason: '${event.reason}'`
      );
      setTimeout(() => {
        log.info(`Attempting to reconenct to WS`);
        this.startWebsocket();
      }, 10000);
    };
    socket.onerror = (event) => {
      log.error(
        `Websocket connection to ${MASTODON_INSTANCE} failed. ${event.type}: ${event.error}, message: '${event.message}'`
      );
    };
  }

  private constructor() {
    this.startWebsocket();
  }

  public static init() {
    if (this._instance === undefined) {
      this._instance = new TimelineReader();
    }
  }

  public static get instance(): TimelineReader {
    TimelineReader.init();
    return this._instance;
  }
}
